package java.util.concurrent;

import java.util.function.Consumer;
import java.util.Spliterators;
import java.io.ObjectInputStream;
import java.io.IOException;
import java.io.ObjectOutputStream;
import java.util.Spliterator;
import java.lang.reflect.Array;
import java.util.NoSuchElementException;
import java.util.Iterator;
import java.util.Collection;
import java.util.concurrent.locks.Condition;
import java.util.concurrent.locks.ReentrantLock;
import java.io.Serializable;
import java.util.AbstractQueue;

public class LinkedBlockingDeque<E> extends AbstractQueue<E> implements BlockingDeque<E>, Serializable
{
    private static final long serialVersionUID = -387911632671998426L;
    transient LinkedBlockingDeque.Node<E> first;
    transient LinkedBlockingDeque.Node<E> last;
    private transient int count;
    private final int capacity;
    final ReentrantLock lock;
    private final Condition notEmpty;
    private final Condition notFull;
    
    public LinkedBlockingDeque() {
        this(Integer.MAX_VALUE);
    }
    
    public LinkedBlockingDeque(int capacity) {
        this.lock = new ReentrantLock();
        this.notEmpty = this.lock.newCondition();
        this.notFull = this.lock.newCondition();
        if (capacity <= 0) {
            throw new IllegalArgumentException();
        }
        this.capacity = capacity;
        return;
    }
    
    public LinkedBlockingDeque(Collection<? extends E> collection) {
        this(Integer.MAX_VALUE);
        ReentrantLock lock = this.lock;
        lock.lock();
        try {
            for (Object next : collection) {
                if (next == null) {
                    throw new NullPointerException();
                }
                if (!this.linkLast(new LinkedBlockingDeque.Node(next))) {
                    throw new IllegalStateException("Deque full");
                }
                continue;
            }
        }
        finally {
            lock.unlock();
        }
    }
    
    private boolean linkFirst(LinkedBlockingDeque.Node<E> last) {
        if (this.count >= this.capacity) {
            return false;
        }
        LinkedBlockingDeque.Node<Object> first = this.first;
        last.next = first;
        this.first = last;
        if (this.last == null) {
            this.last = last;
        }
        else {
            first.prev = last;
        }
        ++this.count;
        this.notEmpty.signal();
        return true;
    }
    
    private boolean linkLast(LinkedBlockingDeque.Node<E> first) {
        if (this.count >= this.capacity) {
            return false;
        }
        LinkedBlockingDeque.Node<Object> last = this.last;
        first.prev = last;
        this.last = first;
        if (this.first == null) {
            this.first = first;
        }
        else {
            last.next = first;
        }
        ++this.count;
        this.notEmpty.signal();
        return true;
    }
    
    private E unlinkFirst() {
        LinkedBlockingDeque.Node<Object> first = this.first;
        if (first == null) {
            return null;
        }
        LinkedBlockingDeque.Node<Object> next = first.next;
        Object item = first.item;
        first.item = null;
        first.next = first;
        if ((this.first = next) == null) {
            this.last = null;
        }
        else {
            next.prev = null;
        }
        --this.count;
        this.notFull.signal();
        return item;
    }
    
    private E unlinkLast() {
        LinkedBlockingDeque.Node<Object> last = this.last;
        if (last == null) {
            return null;
        }
        LinkedBlockingDeque.Node<Object> prev = last.prev;
        Object item = last.item;
        last.item = null;
        last.prev = last;
        if ((this.last = prev) == null) {
            this.first = null;
        }
        else {
            prev.next = null;
        }
        --this.count;
        this.notFull.signal();
        return item;
    }
    
    void unlink(LinkedBlockingDeque.Node<E> node) {
        LinkedBlockingDeque.Node<Object> prev = node.prev;
        LinkedBlockingDeque.Node<Object> next = node.next;
        if (prev == null) {
            this.unlinkFirst();
        }
        else if (next == null) {
            this.unlinkLast();
        }
        else {
            prev.next = next;
            next.prev = prev;
            node.item = null;
            --this.count;
            this.notFull.signal();
        }
    }
    
    @Override
    public void addFirst(E e) {
        if (!this.offerFirst(e)) {
            throw new IllegalStateException("Deque full");
        }
        return;
    }
    
    @Override
    public void addLast(E e) {
        if (!this.offerLast(e)) {
            throw new IllegalStateException("Deque full");
        }
        return;
    }
    
    @Override
    public boolean offerFirst(E e) {
        if (e == null) {
            throw new NullPointerException();
        }
        LinkedBlockingDeque.Node<E> node = new LinkedBlockingDeque.Node(e);
        ReentrantLock lock = this.lock;
        lock.lock();
        try {
            return this.linkFirst(node);
        }
        finally {
            lock.unlock();
        }
    }
    
    @Override
    public boolean offerLast(E e) {
        if (e == null) {
            throw new NullPointerException();
        }
        LinkedBlockingDeque.Node<E> node = new LinkedBlockingDeque.Node(e);
        ReentrantLock lock = this.lock;
        lock.lock();
        try {
            return this.linkLast(node);
        }
        finally {
            lock.unlock();
        }
    }
    
    @Override
    public void putFirst(E e) throws InterruptedException {
        if (e == null) {
            throw new NullPointerException();
        }
        LinkedBlockingDeque.Node<E> node = new LinkedBlockingDeque.Node(e);
        ReentrantLock lock = this.lock;
        lock.lock();
        try {
            while (!this.linkFirst(node)) {
                this.notFull.await();
            }
        }
        finally {
            lock.unlock();
        }
        return;
    }
    
    @Override
    public void putLast(E e) throws InterruptedException {
        if (e == null) {
            throw new NullPointerException();
        }
        LinkedBlockingDeque.Node<E> node = new LinkedBlockingDeque.Node(e);
        ReentrantLock lock = this.lock;
        lock.lock();
        try {
            while (!this.linkLast(node)) {
                this.notFull.await();
            }
        }
        finally {
            lock.unlock();
        }
        return;
    }
    
    @Override
    public boolean offerFirst(E e, long n, TimeUnit timeUnit) throws InterruptedException {
        if (e == null) {
            throw new NullPointerException();
        }
        LinkedBlockingDeque.Node<E> node = new LinkedBlockingDeque.Node(e);
        long n2 = timeUnit.toNanos(n);
        ReentrantLock lock = this.lock;
        lock.lockInterruptibly();
        try {
            while (!this.linkFirst(node)) {
                if (n2 <= 0L) {
                    return false;
                }
                n2 = this.notFull.awaitNanos(n2);
            }
            return true;
        }
        finally {
            lock.unlock();
        }
    }
    
    @Override
    public boolean offerLast(E e, long n, TimeUnit timeUnit) throws InterruptedException {
        if (e == null) {
            throw new NullPointerException();
        }
        LinkedBlockingDeque.Node<E> node = new LinkedBlockingDeque.Node(e);
        long n2 = timeUnit.toNanos(n);
        ReentrantLock lock = this.lock;
        lock.lockInterruptibly();
        try {
            while (!this.linkLast(node)) {
                if (n2 <= 0L) {
                    return false;
                }
                n2 = this.notFull.awaitNanos(n2);
            }
            return true;
        }
        finally {
            lock.unlock();
        }
    }
    
    @Override
    public E removeFirst() {
        Object pollFirst = this.pollFirst();
        if (pollFirst == null) {
            throw new NoSuchElementException();
        }
        return pollFirst;
    }
    
    @Override
    public E removeLast() {
        Object pollLast = this.pollLast();
        if (pollLast == null) {
            throw new NoSuchElementException();
        }
        return pollLast;
    }
    
    @Override
    public E pollFirst() {
        ReentrantLock lock = this.lock;
        lock.lock();
        try {
            return this.unlinkFirst();
        }
        finally {
            lock.unlock();
        }
    }
    
    @Override
    public E pollLast() {
        ReentrantLock lock = this.lock;
        lock.lock();
        try {
            return this.unlinkLast();
        }
        finally {
            lock.unlock();
        }
    }
    
    @Override
    public E takeFirst() throws InterruptedException {
        ReentrantLock lock = this.lock;
        lock.lock();
        try {
            Object unlinkFirst;
            while ((unlinkFirst = this.unlinkFirst()) == null) {
                this.notEmpty.await();
            }
            return unlinkFirst;
        }
        finally {
            lock.unlock();
        }
    }
    
    @Override
    public E takeLast() throws InterruptedException {
        ReentrantLock lock = this.lock;
        lock.lock();
        try {
            Object unlinkLast;
            while ((unlinkLast = this.unlinkLast()) == null) {
                this.notEmpty.await();
            }
            return unlinkLast;
        }
        finally {
            lock.unlock();
        }
    }
    
    @Override
    public E pollFirst(long n, TimeUnit timeUnit) throws InterruptedException {
        long n2 = timeUnit.toNanos(n);
        ReentrantLock lock = this.lock;
        lock.lockInterruptibly();
        try {
            Object unlinkFirst;
            while ((unlinkFirst = this.unlinkFirst()) == null) {
                if (n2 <= 0L) {
                    return null;
                }
                n2 = this.notEmpty.awaitNanos(n2);
            }
            return unlinkFirst;
        }
        finally {
            lock.unlock();
        }
    }
    
    @Override
    public E pollLast(long n, TimeUnit timeUnit) throws InterruptedException {
        long n2 = timeUnit.toNanos(n);
        ReentrantLock lock = this.lock;
        lock.lockInterruptibly();
        try {
            Object unlinkLast;
            while ((unlinkLast = this.unlinkLast()) == null) {
                if (n2 <= 0L) {
                    return null;
                }
                n2 = this.notEmpty.awaitNanos(n2);
            }
            return unlinkLast;
        }
        finally {
            lock.unlock();
        }
    }
    
    @Override
    public E getFirst() {
        Object peekFirst = this.peekFirst();
        if (peekFirst == null) {
            throw new NoSuchElementException();
        }
        return peekFirst;
    }
    
    @Override
    public E getLast() {
        Object peekLast = this.peekLast();
        if (peekLast == null) {
            throw new NoSuchElementException();
        }
        return peekLast;
    }
    
    @Override
    public E peekFirst() {
        ReentrantLock lock = this.lock;
        lock.lock();
        try {
            return (this.first == null) ? null : this.first.item;
        }
        finally {
            lock.unlock();
        }
    }
    
    @Override
    public E peekLast() {
        ReentrantLock lock = this.lock;
        lock.lock();
        try {
            return (this.last == null) ? null : this.last.item;
        }
        finally {
            lock.unlock();
        }
    }
    
    @Override
    public boolean removeFirstOccurrence(Object o) {
        if (o == null) {
            return false;
        }
        ReentrantLock lock = this.lock;
        lock.lock();
        try {
            LinkedBlockingDeque.Node<E> node = this.first;
            while (node != null) {
                if (o.equals(node.item)) {
                    this.unlink(node);
                    return true;
                }
                node = node.next;
            }
            return false;
        }
        finally {
            lock.unlock();
        }
    }
    
    @Override
    public boolean removeLastOccurrence(Object o) {
        if (o == null) {
            return false;
        }
        ReentrantLock lock = this.lock;
        lock.lock();
        try {
            LinkedBlockingDeque.Node<E> node = this.last;
            while (node != null) {
                if (o.equals(node.item)) {
                    this.unlink(node);
                    return true;
                }
                node = node.prev;
            }
            return false;
        }
        finally {
            lock.unlock();
        }
    }
    
    @Override
    public boolean add(E e) {
        this.addLast(e);
        return true;
    }
    
    @Override
    public boolean offer(E e) {
        return this.offerLast(e);
    }
    
    @Override
    public void put(E e) throws InterruptedException {
        this.putLast(e);
    }
    
    @Override
    public boolean offer(E e, long n, TimeUnit timeUnit) throws InterruptedException {
        return this.offerLast(e, n, timeUnit);
    }
    
    @Override
    public E remove() {
        return this.removeFirst();
    }
    
    @Override
    public E poll() {
        return this.pollFirst();
    }
    
    @Override
    public E take() throws InterruptedException {
        return this.takeFirst();
    }
    
    @Override
    public E poll(long n, TimeUnit timeUnit) throws InterruptedException {
        return this.pollFirst(n, timeUnit);
    }
    
    @Override
    public E element() {
        return this.getFirst();
    }
    
    @Override
    public E peek() {
        return this.peekFirst();
    }
    
    @Override
    public int remainingCapacity() {
        ReentrantLock lock = this.lock;
        lock.lock();
        try {
            return this.capacity - this.count;
        }
        finally {
            lock.unlock();
        }
    }
    
    @Override
    public int drainTo(Collection<? super E> collection) {
        return this.drainTo(collection, Integer.MAX_VALUE);
    }
    
    @Override
    public int drainTo(Collection<? super E> collection, int a) {
        if (collection == null) {
            throw new NullPointerException();
        }
        if (collection == this) {
            throw new IllegalArgumentException();
        }
        if (a <= 0) {
            return 0;
        }
        ReentrantLock lock = this.lock;
        lock.lock();
        try {
            int min = Math.min(a, this.count);
            for (int i = 0; i < min; ++i) {
                collection.add(this.first.item);
                this.unlinkFirst();
            }
            return min;
        }
        finally {
            lock.unlock();
        }
    }
    
    @Override
    public void push(E e) {
        this.addFirst(e);
    }
    
    @Override
    public E pop() {
        return this.removeFirst();
    }
    
    @Override
    public boolean remove(Object o) {
        return this.removeFirstOccurrence(o);
    }
    
    @Override
    public int size() {
        ReentrantLock lock = this.lock;
        lock.lock();
        try {
            return this.count;
        }
        finally {
            lock.unlock();
        }
    }
    
    @Override
    public boolean contains(Object o) {
        if (o == null) {
            return false;
        }
        ReentrantLock lock = this.lock;
        lock.lock();
        try {
            LinkedBlockingDeque.Node<Object> node = this.first;
            while (node != null) {
                if (o.equals(node.item)) {
                    return true;
                }
                node = node.next;
            }
            return false;
        }
        finally {
            lock.unlock();
        }
    }
    
    @Override
    public Object[] toArray() {
        ReentrantLock lock = this.lock;
        lock.lock();
        try {
            Object[] array = new Object[this.count];
            int n = 0;
            for (LinkedBlockingDeque.Node node = this.first; node != null; node = node.next) {
                array[n++] = node.item;
            }
            return array;
        }
        finally {
            lock.unlock();
        }
    }
    
    @Override
    public <T> T[] toArray(T[] array) {
        ReentrantLock lock = this.lock;
        lock.lock();
        try {
            if (array.length < this.count) {
                array = (Object[])(Object[])Array.newInstance(array.getClass().getComponentType(), this.count);
            }
            int n = 0;
            for (LinkedBlockingDeque.Node node = this.first; node != null; node = node.next) {
                array[n++] = node.item;
            }
            if (array.length > n) {
                array[n] = null;
            }
            return array;
        }
        finally {
            lock.unlock();
        }
    }
    
    @Override
    public String toString() {
        ReentrantLock lock = this.lock;
        lock.lock();
        try {
            LinkedBlockingDeque.Node<Object> node = this.first;
            if (node == null) {
                return "[]";
            }
            StringBuilder sb = new StringBuilder();
            sb.append('[');
            while (true) {
                Object item = node.item;
                sb.append((item == this) ? "(this Collection)" : item);
                node = node.next;
                if (node == null) {
                    break;
                }
                sb.append(',').append(' ');
            }
            return sb.append(']').toString();
        }
        finally {
            lock.unlock();
        }
    }
    
    @Override
    public void clear() {
        ReentrantLock lock = this.lock;
        lock.lock();
        try {
            LinkedBlockingDeque.Node<Object> next;
            for (LinkedBlockingDeque.Node first = this.first; first != null; first = next) {
                first.item = null;
                next = first.next;
                first.prev = null;
                first.next = null;
            }
            LinkedBlockingDeque.Node<E> node = null;
            this.last = node;
            this.first = node;
            this.count = 0;
            this.notFull.signalAll();
        }
        finally {
            lock.unlock();
        }
    }
    
    @Override
    public Iterator<E> iterator() {
        return new LinkedBlockingDeque.Itr();
    }
    
    @Override
    public Iterator<E> descendingIterator() {
        return new LinkedBlockingDeque.DescendingItr();
    }
    
    @Override
    public Spliterator<E> spliterator() {
        return new LinkedBlockingDeque.LBDSpliterator(this);
    }
    
    private void writeObject(ObjectOutputStream objectOutputStream) throws IOException {
        ReentrantLock lock = this.lock;
        lock.lock();
        try {
            objectOutputStream.defaultWriteObject();
            for (LinkedBlockingDeque.Node node = this.first; node != null; node = node.next) {
                objectOutputStream.writeObject(node.item);
            }
            objectOutputStream.writeObject(null);
        }
        finally {
            lock.unlock();
        }
    }
    
    private void readObject(ObjectInputStream objectInputStream) throws IOException, ClassNotFoundException {
        objectInputStream.defaultReadObject();
        this.count = 0;
        this.first = null;
        this.last = null;
        while (true) {
            Object object = objectInputStream.readObject();
            if (object == null) {
                break;
            }
            this.add(object);
        }
    }
    
    static final class Node<E>
    {
        E item;
        LinkedBlockingDeque.Node<E> prev;
        LinkedBlockingDeque.Node<E> next;
        
        Node(E item) {
            this.item = item;
        }
    }
    
    private abstract class AbstractItr implements Iterator<E>
    {
        LinkedBlockingDeque.Node<E> next;
        E nextItem;
        private LinkedBlockingDeque.Node<E> lastRet;
        
        abstract LinkedBlockingDeque.Node<E> firstNode();
        
        abstract LinkedBlockingDeque.Node<E> nextNode(LinkedBlockingDeque.Node<E> p0);
        
        AbstractItr() {
            ReentrantLock lock = LinkedBlockingDeque.this.lock;
            lock.lock();
            try {
                this.next = this.firstNode();
                this.nextItem = ((this.next == null) ? null : this.next.item);
            }
            finally {
                lock.unlock();
            }
        }
        
        private LinkedBlockingDeque.Node<E> succ(LinkedBlockingDeque.Node<E> node) {
            while (true) {
                LinkedBlockingDeque.Node<E> nextNode = this.nextNode(node);
                if (nextNode == null) {
                    return null;
                }
                if (nextNode.item != null) {
                    return nextNode;
                }
                if (nextNode == node) {
                    return this.firstNode();
                }
                node = nextNode;
            }
        }
        
        void advance() {
            ReentrantLock lock = LinkedBlockingDeque.this.lock;
            lock.lock();
            try {
                this.next = this.succ(this.next);
                this.nextItem = ((this.next == null) ? null : this.next.item);
            }
            finally {
                lock.unlock();
            }
        }
        
        @Override
        public boolean hasNext() {
            return this.next != null;
        }
        
        @Override
        public E next() {
            if (this.next == null) {
                throw new NoSuchElementException();
            }
            this.lastRet = this.next;
            Object nextItem = this.nextItem;
            this.advance();
            return nextItem;
        }
        
        @Override
        public void remove() {
            LinkedBlockingDeque.Node<Object> lastRet = this.lastRet;
            if (lastRet == null) {
                throw new IllegalStateException();
            }
            this.lastRet = null;
            ReentrantLock lock = LinkedBlockingDeque.this.lock;
            lock.lock();
            try {
                if (lastRet.item != null) {
                    LinkedBlockingDeque.this.unlink(lastRet);
                }
            }
            finally {
                lock.unlock();
            }
            return;
        }
    }
    
    private class Itr extends LinkedBlockingDeque.AbstractItr
    {
        @Override
        LinkedBlockingDeque.Node<E> firstNode() {
            return LinkedBlockingDeque.this.first;
        }
        
        @Override
        LinkedBlockingDeque.Node<E> nextNode(LinkedBlockingDeque.Node<E> node) {
            return node.next;
        }
    }
    
    private class DescendingItr extends LinkedBlockingDeque.AbstractItr
    {
        @Override
        LinkedBlockingDeque.Node<E> firstNode() {
            return LinkedBlockingDeque.this.last;
        }
        
        @Override
        LinkedBlockingDeque.Node<E> nextNode(LinkedBlockingDeque.Node<E> node) {
            return node.prev;
        }
    }
    
    static final class LBDSpliterator<E> implements Spliterator<E>
    {
        static final int MAX_BATCH = 33554432;
        final LinkedBlockingDeque<E> queue;
        LinkedBlockingDeque.Node<E> current;
        int batch;
        boolean exhausted;
        long est;
        
        LBDSpliterator(LinkedBlockingDeque<E> queue) {
            this.queue = queue;
            this.est = (long)queue.size();
        }
        
        @Override
        public long estimateSize() {
            return this.est;
        }
        
        @Override
        public Spliterator<E> trySplit() {
            LinkedBlockingDeque<Object> queue = this.queue;
            int batch = this.batch;
            int n = (batch <= 0) ? 1 : ((batch >= 33554432) ? 33554432 : (batch + 1));
            LinkedBlockingDeque.Node<Object> node;
            if (!this.exhausted && ((node = this.current) != null || (node = queue.first) != null) && node.next != null) {
                Object[] array = new Object[n];
                ReentrantLock lock = queue.lock;
                int n2 = 0;
                LinkedBlockingDeque.Node<Object> current = this.current;
                lock.lock();
                try {
                    if (current != null || (current = queue.first) != null) {
                        do {
                            if ((array[n2] = current.item) != null) {
                                ++n2;
                            }
                            else {
                                continue;
                            }
                        } while ((current = current.next) != null && n2 < n);
                    }
                }
                finally {
                    lock.unlock();
                }
                this.current = current;
                if (current == null) {
                    this.est = 0L;
                    this.exhausted = true;
                }
                else {
                    long est = this.est - (long)n2;
                    this.est = est;
                    if (est < 0L) {
                        this.est = 0L;
                    }
                }
                if (n2 > 0) {
                    this.batch = n2;
                    return Spliterators.spliterator(array, 0, n2, 4368);
                }
            }
            return null;
        }
        
        @Override
        public void forEachRemaining(Consumer<? super E> consumer) {
            if (consumer == null) {
                throw new NullPointerException();
            }
            LinkedBlockingDeque<Object> queue = this.queue;
            ReentrantLock lock = queue.lock;
            if (!this.exhausted) {
                this.exhausted = true;
                LinkedBlockingDeque.Node<Object> node = this.current;
                do {
                    Object item = null;
                    lock.lock();
                    try {
                        if (node == null) {
                            node = queue.first;
                        }
                        while (node != null) {
                            item = node.item;
                            node = node.next;
                            if (item != null) {
                                break;
                            }
                            continue;
                        }
                    }
                    finally {
                        lock.unlock();
                    }
                    if (item != null) {
                        consumer.accept(item);
                    }
                    else {
                        continue;
                    }
                } while (node != null);
            }
            return;
        }
        
        @Override
        public boolean tryAdvance(Consumer<? super E> consumer) {
            if (consumer == null) {
                throw new NullPointerException();
            }
            LinkedBlockingDeque<Object> queue = this.queue;
            ReentrantLock lock = queue.lock;
            if (!this.exhausted) {
                Object item = null;
                lock.lock();
                try {
                    if (this.current == null) {
                        this.current = queue.first;
                    }
                    while (this.current != null) {
                        item = this.current.item;
                        this.current = this.current.next;
                        if (item != null) {
                            break;
                        }
                        continue;
                    }
                }
                finally {
                    lock.unlock();
                }
                if (this.current == null) {
                    this.exhausted = true;
                }
                if (item != null) {
                    consumer.accept(item);
                    return true;
                }
            }
            return false;
        }
        
        @Override
        public int characteristics() {
            return 4368;
        }
    }
}
